from .base import (
    Any,
    Boolean,
    Const,
    Integer,
    List,
    Null,
    Object,
    OneOf,
    Optional,
    Reference,
    String,
    Type,
    Unset)

from .validation import validate
from .json_schema import json_schema
from .typescript import to_ts
from .query_arguments import parse_query_argument, QueryValidationError




def parse_query_argument(type_, value, values, referencetypes={}):
    return type_.visit(Visitor(referencetypes), value, values)


class QueryValidationError(Exception):
    pass


class Visitor:

    def __init__(self, types):
        self.types = types

    def visit_string(self, node, value, values):
        return value

    def visit_integer(self, node, value, values):
        try:
            return int(value)
        except ValueError:
            raise QueryValidationError(f"Invalid value for integer: { value }")

    def visit_list(self, node, value, values):
        return [node.type.visit(self, item, [item]) for item in values]

    def visit_default(self, typename, node, argument_name, request):
        raise Exception(f"Invalid Type for argument parsing ({ typename })")

    def visit_reference(self, node, value, values):
        reftype = self.types.get(node.name)
        if reftype is None:
            raise Exception(f"Unknown type { node.name }")
        return reftype.visit(self, value, values)

    def visit_const(self, node, value, values):
        if isinstance(node.value, int):
            try:
                value = int(value)
            except ValueError:
                raise QueryValidationError(f"Invalid const { repr(value) }, "
                                            f"expected { repr(node.value )}")
        if node.value != value:
            raise QueryValidationError(f"Invalid const { repr(value) }, "
                                        f"expected { repr(node.value )}")
        return value

    def visit_oneof(self, node, value, values):
        for subtype in node.types:
            try:
                return subtype.visit(self, value, values)
            except QueryValidationError:
                continue
        raise QueryValidationError(f"Invalid value { repr(value) }")



from .base import Unset


def validate(type_, value, types):
    """ validates the given value against the type

    uses the given types to resolve references if needed.

    >>> validate(Object({"msg": String()}), {"msg": 23}, {})
    <<< [('msg',), 'should be String']

    :param type_:
        The type to validate against
    :param value:
        The value to validate
    :param types:
        dictionary of types, to be used to resolve references in `type_`.
    :returns:
        list of (path, message) tuples describing the validation errors.
        Empty, if validation was successful
    """
    return list(type_.visit(Visitor(types), value))


class Visitor:

    def __init__(self, types):
        self.types = types

    def visit_any(self, node, value):
        return []

    def visit_null(self, node, value):
        if value is not None:
            yield (), 'should be null'

    def visit_string(self, node, value):
        if not isinstance(value, str):
            yield (), 'should be a string'

    def visit_integer(self, node, value):
        if not isinstance(value, int):
            yield (), 'should be an integer'

    def visit_boolean(self, node, value):
        if not isinstance(value, bool):
            yield (), 'should be a boolean'

    def visit_object(self, node, value):
        if not isinstance(value, dict):
            yield (), 'should be an object'
            return

        for name, type_ in node.fields.items():
            if name in node.optional_keys and name not in value:
                continue
            elif name not in value:
                yield (name,), "value missing"
            else:
                for path, msg in type_.visit(self, value[name]):
                    yield (name, *path), msg

    def visit_list(self, node, value):
        if not isinstance(value, list):
            yield (), 'should be a list'
            return

        for index, item in enumerate(value):
            for path, msg in node.type.visit(self, item):
                yield (index, *path), msg

    def visit_const(self, node, value):
        if not node.value == value:
            yield ((), 'should be %r' % node.value)

    def visit_unset(self, node, value):
        if not value == Unset.UNSET:
            yield ((), 'should be unset')

    def visit_reference(self, node, value):
        type_ = self.types[node.name]
        return type_.visit(self, value)

    def visit_oneof(self, node, value):
        if not any(not list(type_.visit(self, value)) for type_ in node.types):
            text = ", ".join(str(type_) for type_ in node.types)
            yield ((), f'must be one of { text }')

    def visit_default(self, name, node):
        raise Exception(f"cannot handle type { name }")




def json_schema(type_):
    """ generates a json schema from the given type

    >>> json_schema(String())
    <<< {'type': 'null'}
    """
    return type_.visit(Visitor())


class Visitor:

    def visit_any(self, node):
        return {}

    def visit_null(self, node):
        return {
            "type": "null"
        }

    def visit_string(self, node):
        return {
            "type": "string"
        }

    def visit_integer(self, node):
        return {
            "type": "integer"
        }

    def visit_boolean(self, node):
        return {
            "type": "boolean"
        }

    def visit_object(self, node):
        return {
            "type": "object",
            "properties": {name: field.visit(self)
                    for name, field in node.fields.items()},
            "required": list(node.fields.keys()),
            "additionalProperties": False,
        }

    def visit_list(self, node):
        return {
            "type": "array",
            "items": node.type.visit(self)
        }

    def visit_const(self, node):
        return {
            "enum": [node.value]
        }

    def visit_unset(self, node):
        return {"not": {}}

    def visit_reference(self, node):
        return {"$ref": f"#/components/schemas/{ node.name }"}

    def visit_default(self, name, node):
        raise Exception(f"cannot handle type { name }")



from textwrap import dedent
import re

from jinja2 import Environment, StrictUndefined

from .types import to_ts


env = Environment()
env.filters['has_body'] = lambda value: value in ['post', 'patch', 'put']
env.filters['to_ts'] = to_ts
env.filters['path_to_placeholder'] = lambda value: re.sub(
        '\\{([ a-zA-Z_]+)\\}', lambda m: '${ ' + m.groups()[0] + ' }', value)

env.undefined = StrictUndefined


api_template = env.from_string(dedent("""
    import { Observable, from } from 'rxjs';
    import { map, concatMap } from 'rxjs/operators';
    import { ajax } from 'rxjs/ajax';
    import { fromFetch } from 'rxjs/fetch';

    type Param = string | number | undefined | null;

    function encodeParams(params: {[name: string]: Param}): string {
        return Object
            .keys(params)
            .filter(key => params[key] !== null && params[key] !== undefined)
            .map(key => `${ key }=${ encodeURIComponent(params[key]!) }`)
            .join('&');
    }

    {% for name, type in styx.types.items() %}
    export type {{ name }} = {{ type | to_ts }};
    {% endfor %}

    export class Api {

        protected request(method: string, url: string, body?: any): Observable<any> {
          const options = {
            method: method.toUpperCase(),
            ...(body !== undefined
                ? {
                  headers: { 'Content-Type': 'application/json' },
                  body: JSON.stringify(body)
                }
                : {}
            )
          }

          return fromFetch(new Request(url, options)).pipe(
            concatMap(r => from(r.json())));
        }

        protected requestX(method: string, url: string, body?: any): Observable<any> {
            if (body === undefined) {
                return ajax({ url, method: method.toUpperCase() })
                    .pipe(map(r => r.response));
            }
            else {
                return ajax({
                    url,
                    method: method.toUpperCase(),
                    headers: { 'Content-Type': 'application/json' },
                    body: JSON.stringify(body)
                }).pipe(map(r => r.response));;
            }
        }

    {% for operation in styx.operations %}
        public {{ operation.operation_id }}(
            {%- for param in operation.parameters -%}
                {{ param.name }}: any,
            {%- endfor -%}
            {%- if operation.method | has_body -%}
                body: {{ operation.body_type | to_ts }}
            {%- endif -%}
        ): Observable<{{ operation.response_type | to_ts }}> {
            const url = `{{ operation.path | path_to_placeholder }}?${ encodeParams({
                {%- for param in operation.parameters -%}
                    {%- if param.in == 'query' -%}
                        {{ param.name }},
                    {%- endif -%}
                {%- endfor -%}
            }) }`;
            {%- if operation.method | has_body %}
            return this.request({{ operation.method | tojson }}, url, body);
            {% else %}
            return this.request({{ operation.method | tojson }}, url);
            {%- endif %}
        }
    {% endfor %}
    }

    export const api = new Api();

"""))




class Type:
    def visit(self, visitor, *args):
        name = type(self).__name__.lower()
        func = getattr(visitor, f'visit_{ name }', None)
        if func:
            return func(self, *args)
        else:
            return visitor.visit_default(name, self, *args)

    def __str__(self):
        return type(self).__name__


class String(Type):
    pass


class Integer(Type):
    pass


class Boolean(Type):
    pass


class Null(Type):
    pass


class Object(Type):
    def __init__(self, fields):
        self.optional_keys = frozenset(key for key, value
                                in fields.items()
                                if isinstance(value, Optional))

        self.fields = {key: value.type if isinstance(value, Optional) else value
                        for key, value in fields.items()}


class Optional:
    """ This is NOT a type
    """
    def __init__(self, type_):
        self.type = type_


class List(Type):
    def __init__(self, type_):
        self.type = type_


class Any(Type):
    pass


class Const(Type):
    def __init__(self, value):
        self.value = value


class Reference(Type):
    def __init__(self, name):
        self.name = name


class OneOf(Type):
    def __init__(self, *types):
        self.types = types


class Unset(Type):
    class UNSET:
        pass



import os
from itertools import chain

from setuptools import setup, Command, find_packages


def packages_for(*names):
    return sorted(list(names) + list(chain.from_iterable(
        ['{}.{}'.format(name, p) for p in find_packages(name)]
        for name in names)))


setup(
    name='styxapi',
    version='3.2.0',
    packages=packages_for('styx'),
    install_requires=[
        "Flask>=1.0.4",
        "pyyaml>=5.4.1",
    ],
    extras_require={
        'dev': [
            'pycodestyle==2.7.0',
            'pytest==6.2.4',
            'pylint==2.9.6',
        ]
    },
    entry_points={}
)



import json


def to_ts(type_):
    """ Generates a typescript type from the given type

    >>> to_ts(Array(String()))
    <<< "string[]"
    """
    return type_.visit(Visitor())


class Visitor:

    def visit_unset(self, node):
        return 'undefined'

    def visit_any(self, node):
        return 'any'

    def visit_null(self, node):
        return 'null'

    def visit_string(self, node):
        return 'string'

    def visit_integer(self, node):
        return 'number'

    def visit_boolean(self, node):
        return 'boolean'

    def visit_object(self, node):
        fields = [
            f'"{ name }"{ "?" if name in node.optional_keys else "" }: '
            f'({ field.visit(self) })'
            for name, field in node.fields.items()]

        return '{' + ', '.join(fields) + '}'

    def visit_list(self, node):
        return f'({ node.type.visit(self) })[]'

    def visit_const(self, node):
        return json.dumps(node.value)

    def visit_reference(self, node):
        return node.name

    def visit_oneof(self, node):
        if node.types:
            return ' | '.join(f'({ t.visit(self) })' for t in node.types)
        else:
            return 'never'

    def visit_default(self, name, node):
        raise Exception(f"cannot handle type { name }")



import re
import json
import yaml
from functools import wraps
from inspect import signature, Parameter

import click
from flask import request, jsonify, abort

from .types import (validate, json_schema, parse_query_argument,
                        QueryValidationError, Reference, String)
from .generation import api_template


class SchemaValidationError(Exception):
    pass


class Operation:
    def __init__(self, operation_id, summary, path, method,
                response_type, parameters, body_type, types):
        self.operation_id = operation_id
        self.path = path
        self.summary = summary
        self.method = method
        self.response_type = response_type
        self.parameters = parameters
        self.body_type = body_type
        self.types = types

    def assert_function_fit(self, func):
        def parameter_fits(p, name, optional):
            if name == p.name and p.kind in [Parameter.POSITIONAL_OR_KEYWORD,
                                                Parameter.KEYWORD_ONLY]:
                return optional or p.default is not Parameter.empty
            elif p.kind == Parameter.VAR_KEYWORD:
                return True
            else:
                return False

        sig = signature(func, follow_wrapped=False)
        for desc in self.parameters:
            assert any(parameter_fits(p, desc['name'], desc['required'])
                                            for p in sig.parameters.values()), \
                f'callable does not handle parameter { desc["name"] }'

    def get_parameters(self, path_params, req):
        kwargs = {}

        for desc in self.parameters:
            name = desc['name']

            if desc['in'] == 'query':
                if name in req.args:
                    try:
                        kwargs[name] = parse_query_argument(desc['type'],
                                                        req.args[name],
                                                        req.args.getlist(name),
                                                        self.types)
                    except QueryValidationError as exc:
                        raise SchemaValidationError(str(exc))
                elif desc['required']:
                    raise SchemaValidationError(f'parameter { desc["name"] } is required')
            elif desc['in'] == 'path':
                kwargs[name] = parse_query_argument(desc['type'],
                                                        path_params[name],
                                                        [],
                                                        self.types)
            else:
                raise Exception("unknown 'in' value")

        if self.body_type:
            errors = list(validate(self.body_type, request.json, self.types))
            if errors:
                raise SchemaValidationError(
                    'Malformed body\n\n%s' % _format_errors(errors))
            kwargs = {"body": request.json, **kwargs}

        return kwargs


class Styx:
    def __init__(self, app, version, prefix=''):
        self.app = app
        self.version = version
        self.prefix = prefix.rstrip('/')

        self.types = {}
        self.operations = []

        @app.cli.command()
        @click.argument('file', type=click.File('w'), default='-')
        def generate_client(file):
            """ Generates a typescript client
            """
            file.write(api_template.render(styx=self))

    def define_type(self, name, type_):
        self.types[name] = type_
        return Reference(name)

    def operation(self, operation_id, *,
                            summary,
                            path,
                            method,
                            response_type,
                            parameters=None,
                            body_type=None):

        assert path.startswith('/')

        self.operations.append(
            Operation(operation_id, summary, self.prefix + path, method,
                        response_type, parameters or [], body_type, self.types))

    def openapi_spec(self, server_url):
        paths = {}

        for op in self.operations:
            methods = paths.setdefault(op.path, {})
            methods[op.method] = item = {
                "operationId": op.operation_id,
                "summary": op.summary,
                "parameters": op.parameters,
                "responses": {
                    "200": {
                        "description": "Successful response",
                        "content": {
                            "application/json": {
                                "schema": json_schema(op.response_type)
                            }
                        }
                    }
                }
            }
            if op.body_type:
                item['requestBody'] = {
                    "content": {
                        "application/json": {
                            "schema": json_schema(op.body_type)
                        }
                    }
                }

        return {
            'openapi': '3.0.3',
            'info': {
                'title': 'Example API',
                'version': self.version,
            },
            'servers': [{
                "url": server_url,
                "description": "this server",
            }],
            'components': {
                'schemas': {
                    name: json_schema(value) for name, value in self.types.items()
                }
            },
            'paths': paths
        }

    def register_apispec(self, path="/openapi/"):
        @self.app.route(path)
        def apispec():
            return jsonify(self.openapi_spec(request.base_url))

    def register_handler(self, name=None):
        def decorator(func):
            operation = self._find_operation(name or func.__name__)
            operation.assert_function_fit(func)

            path = re.sub(
                        '[{]([^}]*)[}]',
                        lambda m: f'<string:{ m.groups()[0] }>',
                        operation.path)

            @self.app.route(path, methods=[operation.method.upper()])
            @wraps(func)
            def wrapper(**path_params):
                try:
                    kwargs = operation.get_parameters(path_params, request)
                except SchemaValidationError as exc:
                    self.app.logger.info("request validation failed: %r", exc)
                    abort(400)

                response = func(**kwargs)

                errors = validate(operation.response_type, response, self.types)
                if not errors:
                    return response
                else:
                    raise SchemaValidationError(_format_errors(errors)
                                    + '\n\n' + json.dumps(response))

            return func
        return decorator

    def _find_operation(self, name):
        return next(op
                for op in self.operations
                if op.operation_id == name)

    def get_yaml_spec(self, server_url):
        return yaml.dump(self.openapi_spec(server_url))


def _format_errors(errors):
    return '\n'.join(f'.{ ".".join(repr(p) for p in path) } { msg }'
                    for path, msg in errors)


def query_parameter(name, description, type_=String(), required=False):
    return {
        "name": name,
        "in": "query",
        "description": description,
        "required": required,
        "schema": json_schema(type_),

        "type": type_,
    }


def path_parameter(name, description, type_=String()):
    return {
        "name": name,
        "in": "path",
        "description": description,
        "required": True,
        "schema": json_schema(type_),
        "type": type_,
    }
